#include "offsets.h"

extern task_attributes task_map[];
extern message_attributes message_map[];
extern trans_attributes trans_map[];
extern int num_messages;
extern int num_trans;
extern timeval max_resp;
extern int tractable_test;

static int find_corresp_r();
static void findmaxW();
static timeval findw();
static timeval wnplus();
static timeval I();
static void firstWvalue();
static int nextWvalue();
static int permuteWvalue();
static timeval alltransI();
static int K();
static int v();

extern int max();
extern int ceildiv();
extern int floordiv();

timeval find_arrival_time(i)
int i;
{
	int q;							  /* defined by the analysis */
	int t;							  /* the transaction of which task i is a member */
	int must_continue;			  /* set to false when termination condition met */
	int j;
	int source_cpu;
	int not_remote;

	not_remote = TRUE;
	source_cpu = task_map[message_map[i].source].host;
	for (j = 0; ((j < message_map[i].num_dests) && (not_remote)) ; j++) {
		if (source_cpu != task_map[message_map[i].dest[j]].host)
			not_remote = FALSE;
	}
	if (not_remote)
		return 0;

	/* see technical note: all we need to do here is find w over significant
	 * busy periods
	 */
	message_map[i].r = 0;
	t = message_map[i].trans;
	q = 0;
	do {
		int x;

		must_continue = FALSE;
		if (tractable_test) {
			firstWvalue(t);
			do {
				must_continue = find_corresp_r(i, t, q);
			}
			while (nextWvalue(t));
		} else {
			for (x = 0; x < num_trans; x++) {
				firstWvalue(x);
			}
			/* iterate over all messages in transaction t */
			do {
				must_continue = find_corresp_r(i, t, q);
			}
			while (permuteWvalue(0));
		}
		q++;
	}
	while (must_continue);

	/* ri cannot be less than Ci */
	if (message_map[i].r < message_map[i].C) {
		fprintf(stderr,
				  "assertion failure: r(%d) is < Ci(%d) (%d)\n",
				  i, message_map[i].C, message_map[i].r);
	}

	return message_map[i].r;
}

void set_net_Bi()
{
	int i, j;
	
	for (i = 0; i < num_messages; i++) {
		message_map[i].B = 0;
		for (j = 0; j < num_messages; j++) {
			if ((message_map[j].priority > message_map[i].priority) && (message_map[j].trans != message_map[i].trans)) {
				message_map[i].B = max(message_map[i].B, message_map[j].C);
			}
		}
	}
}

static int find_corresp_r(i, t, q)
int i, t, q;
{
	int kj;
	timeval Oi, Ji, Wt, Tt;
	int ei;
	timeval tmp_ri;
	timeval w;

	Oi = message_map[i].O;
	Ji = message_map[i].J;
	ei = message_map[i].e;
	Tt = trans_map[t].T;

	Wt = trans_map[t].W;
	kj = ceildiv(Wt - Oi - Ji, Tt);
	w = findw(i, q);
	tmp_ri = w + Wt - Tt * (q * ei + kj) - Oi;
	message_map[i].r = max(message_map[i].r, tmp_ri);

	/* if the busy period finishes after the latest release of i
	 * then we need to look further
	 */
	if (w + Wt > Tt * ((q + 1) * ei + kj) + Oi) {
		return 1;
	} else {
		return 0;
	}
}

/* iterates to find a value for w; will return w=max_resp if w exceeds
 * max_resp (this is needed to cut short overly long iterations when the
 * utilisation is close to 100%)
 *
 * i is the message in "level i busy period", q is defined by the analysis
 * and Wt is defined by the analysis
 */
static timeval findw(i, q)
int i, q;
{
	timeval w, wnp;
	timeval Ci = message_map[i].C;

	w = Ci;
	wnp = wnplus(i, q, w);
	while (w != wnp && w < max_resp) {
		w = wnp;
		wnp = wnplus(i, q, w);
	}

	if (w > max_resp) {
		w = max_resp;
	}

	return w;
}


/* return w(n+1): see technical note for details
 * takes variable i, q value, and Wt value
 */
static timeval wnplus(i, q, w)
int i, q;
timeval w;
{
	timeval Ci, Bi;
	timeval sum;
	int t;

	Ci = message_map[i].C;
	Bi = message_map[i].B;

	t = message_map[i].trans;

	sum = Ci + q * Ci + Bi;

	/* sum up over this transaction */
	sum += I(t, i, w);

	/* sum up over the rest of the transactions */
	sum += alltransI(t, i, w);

	return sum;
}

/* find interference from all transactions other than t over level
 * i busy period of width w
 */
static timeval alltransI(t, i, w)
timeval w;
int i, t;
{
	timeval sum = 0;
	int x;

	for (x = 0; x < num_trans; x++) {
		if (x != t) {
			/* only need to find the value of W if using tractable test */
			if (tractable_test) {
				findmaxW(x, i, w);  /* set Wx to give largest interference */
			}
			sum += I(x, i, w);
		}
	}
	return sum;
}

/* set W of transaction t to first value */
static void firstWvalue(t)
{
	int k;

	if (trans_map[t].num_messages == 0) {
		fprintf(stderr,
				  "error: transaction %s has no member messages!\n",
				  trans_map[t].ID);
		return;
	}
	trans_map[t].message_num = 0;
	k = trans_map[t].message_map[0];
	trans_map[t].W = message_map[k].O + message_map[k].J;
}

/* set W of transaction t to the next value, return 1 if more, 0 if reached end
*/
static int nextWvalue(t)
{
	int k;

	trans_map[t].message_num++;
	if (trans_map[t].message_num == trans_map[t].num_messages) {
		return 0;
	}
	k = trans_map[t].message_map[trans_map[t].message_num];
	trans_map[t].W = message_map[k].O + message_map[k].J;

	return 1;
}

/* find another W permutation for transaction x, not including transaction
 * t; return 1 if still more permutations, 0 if no more.
 */
static int permuteWvalue(x)
int x;
{
	if (!nextWvalue(x)) {		  /* wrap around */
		firstWvalue(x);
		x++;
		if (x >= num_trans) {
			return 0;				  /* finished the lot */
		} else {
			return permuteWvalue(x);
		}
	} else {
		return 1;
	}
}

/* set the value of W for transaction t that gives maximal interference 
 * over w for messages of higher priority than message i
 */
static void findmaxW(t, i, w)
int t, i;
timeval w;
{
	timeval maxW;
	timeval maxI;
	int x;

	maxI = 0;
	maxW = 0;
	for (x = 0; x < trans_map[t].num_messages; x++) {
		int j;
		int Oj, Jj;
		int curI;

		j = trans_map[t].message_map[x];

		Oj = message_map[j].O;
		Jj = message_map[j].J;

		trans_map[t].W = Oj + Jj;
		curI = I(t, i, w);
		if (curI > maxI) {
			maxI = curI;
			maxW = trans_map[t].W;
		}
	}
	trans_map[t].W = maxW;
}

/* return the total interference from transaction t on a level i busy
 * period of width w
 */
static timeval I(t, i, w)
int t, i;
timeval w;
{
	int sum;
	int x;

	sum = 0;
	for (x = 0; x < trans_map[t].num_messages; x++) {
		int j;

		j = trans_map[t].message_map[x];

		if (message_map[j].priority < message_map[i].priority
			 && message_map[j].net == message_map[i].net) {
			int Cj;

			Cj = message_map[j].C;
			sum += K(j, t, w) * Cj;
		}
	}

	return sum;
}

/* defined by the analysis; j is a member of transaction t */
static int K(j, t, w)
int j, t;
timeval w;
{
	timeval Tt, Wt, Oj;
	int ej;

	Tt = trans_map[t].T;
	Wt = trans_map[t].W;
	Oj = message_map[j].O;
	ej = message_map[j].e;

	return ceildiv(Wt + w - Oj - v(j, t) * Tt, ej * Tt);
}

/* define by analysis; j is a member of transaction t */
static int v(j, t)
int j, t;
{
	timeval Wt, Tt, Oj, Jj;

	Wt = trans_map[t].W;
	Tt = trans_map[t].T;
	Oj = message_map[j].O;
	Jj = message_map[j].J;

	return ceildiv(Wt - Oj - Jj, Tt);
}
