#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAX_LINE    200
#define MAX_STRING  64
#define MAX_TASKS   400
#define MAX_SEMS    400
#define MAX_LOCKS   400
#define MAX_TRANS   100
#define MAX_PRECS   100
#define MAX_EXCLS   100
#define MAX_MESSAGES 400
#define MAX_CPUS    40
#define MAX_NETS    1
#define TRUE        1
#define FALSE       0
#define LARGE_INT   10000000

typedef int timeval;

typedef struct trans_attributes {
	char ID[MAX_STRING];
	timeval T;
	int num_tasks;			/* number of tasks in the transaction */
	int num_messages;
	int task_map[MAX_TASKS];	/* map of tasks contained in the transaction */
	int message_map[MAX_MESSAGES];
	timeval W;			/* as defined by the analysis */
	int task_num;			/* index into task_map[] for which W is set to O+J */
	int message_num;
} trans_attributes;

typedef struct task_attributes {
	timeval D;
	timeval C;
	timeval B;
	timeval J;
	timeval O;
	int e;
	timeval T;			/* calculated from transaction period and e */
	int trans;			/* transaction of which the task is a member */
	int host;			/* CPU on which this task is located, index to cpu_map */
	char ID[MAX_STRING];
	int priority;			/* 0 = lowest, num_tasks-1 = highest */
	timeval oldr;
	timeval r;
	timeval dmsr;
	timeval remaining;
	timeval maxfoundri;
	timeval runat;
	int runnable;
	int blocking_override;
	int receives;
} task_attributes;

typedef struct sem_attributes {
	char ID[MAX_STRING];
	int ceiling;
} sem_attributes;

typedef struct lock_attributes {
	char ID[MAX_STRING];
	int sem;			/* the semaphore being locked - index into table */
	int task;			/* the task making the lock - index into table */
	timeval C;			/* worst-case execution time spent holding the lock */
} lock_attributes;

typedef struct prec_attributes {
	int a;				/* tasks a and b in precedence relation */
	int b;
	int result;			/* boolean result of constraint being met or not */
} prec_attributes;

typedef struct excl_attributes {
	int a;				/* tasks a and b in exclusion relation */
	int b;
	int result;
} excl_attributes;

typedef struct message_attributes {
	timeval D;
	timeval C;
	timeval B;
	timeval J;
	timeval O;
	int e;
	timeval T;			/* calculated from transaction period and e */
	int trans;			/* transaction of which the task is a member */
	int net;			/* network on which this task is located, index to net_map */
	char ID[MAX_STRING];
	int priority;			/* 0 = lowest, num_tasks-1 = highest */
	int source;			/* index into task_map */
	int num_dests;			/* number of destination tasks */
	int dest[MAX_TASKS];		/* index into task_map */
	timeval r;
	timeval oldr;			/* for use in making the outer iteration */
} message_attributes;

typedef struct cpu_attributes {
	int task_map[MAX_TASKS];	/* list of tasks located on this processor */
	char ID[MAX_STRING];		/* host ID */
} cpu_attributes;

typedef struct net_attributes {
	int mess_map[MAX_TASKS];	/* list of messages located on this processor */
	char ID[MAX_STRING];		/* network ID */
	char type[MAX_STRING];		/* network type */
} net_attributes;
